/*
 * This file is part of John the Ripper password cracker,
 * Copyright (c) 2013 by Solar Designer
 * Copyright (c) 2013 by magnum
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted.
 *
 * There's ABSOLUTELY NO WARRANTY, express or implied.
 */

#include <stdio.h> /* for fprintf(stderr, ...) */
//#include <string.h> /* for isalpha(), isxdigit() */
#include<string.h>
#include<stdlib.h>

#include "misc.h" /* for error() */
#include "logger.h"
#include "recovery.h"
#include "os.h"
#include "signals.h"
#include "status.h"
#include "options.h"
#include "rpp.h"
#include "external.h"
#include "cracker.h"
#include "john.h"
#include "mask.h"

#include "memdbg.h"

#define MAX_NUM_MASK_PLHDR 128 // Maximum number of placeholder in a mask.

typedef struct {
	int stack_op_br[MAX_NUM_MASK_PLHDR]; // store location of op braces in mask
	int stack_cl_br[MAX_NUM_MASK_PLHDR]; // store location of cl braces in mask
	int stack_qtn[MAX_NUM_MASK_PLHDR]; // store location of valid ? in mask
	int parse_ok; // 1 if parse is successful, otherwise 0
} parsed_ctx;

 /* Range of characters for a placeholder in the mask */
typedef struct {
	/* Characters in the range */
	unsigned char chars[0x100];
	/* Set to zero when the characters in the range are not consecutive,
	 * otherwise start is set to the minumum value in range. Minimum
	 * value cannot be a null character which has a value zero.
	 */
	unsigned char start;
	/* Number of characters in the range */
	int count;
	/* Postion of the characters in key */
	int pos;
	/* current postion in chars while iterating */
	char iter;
	/* next active range */
	int next;
} mask_range;

/* Processed mask structure for password generation on CPU */
typedef struct {
	/* Set of mask placeholders for generating password
	 */
	mask_range ranges[MAX_NUM_MASK_PLHDR];
	/* Positions in mask for iteration on CPU
	 */
	int active_positions[MAX_NUM_MASK_PLHDR];
	/* Number of postions for iterating on CPU
	 */
	int count;
	/* offset at which mask starts in the key */
	int offset;

} cpu_mask_context;

parsed_ctx parsed_mask;
cpu_mask_context cpu_mask_ctx;

static struct rpp_context ctx, rec_ctx;
static double cand;

/*
 * A "sequence number" for distributing the candidate passwords across nodes.
 * It is OK if this number overflows once in a while, as long as this happens
 * in the same way for all nodes (must be same size unsigned integer type).
 */
static unsigned int seq, rec_seq;

#define store_op(k, i) \
	parsed_mask->stack_op_br[k] = i;

#define store_cl(k, i) \
	parsed_mask->stack_cl_br[k] = i;

#define load_op(i) \
	parsed_mask->stack_op_br[i]

#define load_cl(i) \
	parsed_mask->stack_cl_br[i]

#define load_qtn(i) \
	parsed_mask->stack_qtn[i]

/* valid braces:
 * [abcd], [[[[[abcde], []]abcde]]], [[[ab]cdefr]]
 * invalid braces:
 * [[ab][c], parsed as two separate ranges [[ab] and [c]
 * [[ab][, error, sets parse_ok to 0.
 */
void parse_braces(char *mask, parsed_ctx *parsed_mask) {

	int i, j ,k;
	int cl_br_enc;

	for (i = 0; i < MAX_NUM_MASK_PLHDR; i++) {
		store_cl(i, -1);
		store_op(i, -1);
	}

	j = k = 0;
	while (j < strlen(mask)) {

		for (i = j; i < strlen(mask); i++)
			if (mask[i] == '[')
				break;

		if (i < strlen(mask))
			store_op(k, i); // store first opening brace for kth placeholder

		i++;

		cl_br_enc = 0;
		for (;i < strlen(mask); i++) {
			if (mask[i] == ']') {
				store_cl(k, i); // store last closing brace for kth placeholder
				cl_br_enc = 1;
			}
			if (mask[i] == '[' && cl_br_enc)
				break;
		}

		j = i;
		k++;
	}

	parsed_mask->parse_ok = 1;
	for (i = 0; i < MAX_NUM_MASK_PLHDR; i++)
		if ((load_op(i) == -1) ^ (load_cl(i) == -1))
			parsed_mask->parse_ok = 0;
}

/* Stores the valid ? placeholders in a stack_qtn
 * valid:
 * if outside [] braces and
 * if ? is immediately followed by the identifier such as ?a for all printable ASCII.
 */
void parse_qtn(char *mask, parsed_ctx *parsed_mask) {
	int i, j, k;

	for (i = 0; i < MAX_NUM_MASK_PLHDR; i++)
		parsed_mask->stack_qtn[i] = -1;

	for (i = 0, k = 0; i < strlen(mask); i++) {
		if (mask[i] == '?')
			if (i + 1 < strlen(mask))
				if (isalpha((int)mask[i + 1])) {
					j = 0;
					while (load_op(j) != -1 && load_cl(j) != -1) {
						if (i > load_op(j) && i < load_cl(j))
							goto cont;
						j++;
					}
					parsed_mask->stack_qtn[k++] = i;
				}

		cont:;
	}
}

/* Maps the postion of a range in a mask to its actual postion in a key.
 * Offset for wordlist + mask is not taken into account.
 */
int calc_pos_in_key(char *mask, parsed_ctx *parsed_mask, int mask_loc) {
	int i, ret_pos;

	int search_qtn(parsed_ctx *parsed_mask, int j) {
		int t;
		for (t = 0; load_qtn(t) != -1; t++)
			if(load_qtn(t) == j)
				return t;
		return -1;
		}

	int search_op_stack(parsed_ctx *parsed_mask, int j) {
		int t;
		for (t = 0; load_op(t) != -1; t++)
			if(load_op(t) <= j && load_cl(t) >= j)
				return t;
		return -1;
		}

	i = ret_pos = 0;
	while (i < mask_loc) {
		int t;

		if (search_qtn(parsed_mask, i) != -1)
			i = i + 2;

		else if ((t = search_op_stack(parsed_mask, i)) != -1)
			i = load_cl(t) + 1;

		else
			i++;
		ret_pos++;
	}

	return ret_pos;
}

static void init_cpu_mask(char *mask, parsed_ctx *parsed_mask, cpu_mask_context *cpu_mask_ctx) {
	int i, qtn_ctr, op_ctr, cl_ctr ;
#define count(i) cpu_mask_ctx->ranges[i].count
	for (i = 0; i < MAX_NUM_MASK_PLHDR; i++) {
		cpu_mask_ctx->ranges[i].start =
		cpu_mask_ctx->ranges[i].count =
		cpu_mask_ctx->ranges[i].pos =
		cpu_mask_ctx->ranges[i].iter =
		cpu_mask_ctx->active_positions[i] = 0;
		cpu_mask_ctx->ranges[i].next = MAX_NUM_MASK_PLHDR - 1;
	}
	cpu_mask_ctx->count = cpu_mask_ctx->offset = 0;

	qtn_ctr = op_ctr = cl_ctr = 0;
	for (i = 0; i < MAX_NUM_MASK_PLHDR; i++) {
		if ((unsigned int)load_op(op_ctr) < (unsigned int)load_qtn(qtn_ctr)) {
			int j, k, start_c[50], end_c[50], dash_loc_beg[50], dash_loc_end[50];

			for (j = 0; j < 50; j++)
				start_c[j] = end_c[j] = dash_loc_beg[j] = dash_loc_end[j] = -1;

			cpu_mask_ctx->ranges[i].pos = calc_pos_in_key(mask, parsed_mask, load_op(op_ctr));

			for (j = 0, k = 0; j < load_cl(cl_ctr); j++) {
				if (mask[j] == '-')
					if (j - 1 > load_op(op_ctr))
						if (mask[j - 1] != '-')
							if (j + 1 < load_cl(cl_ctr)) {
								if (j + 4 < load_cl(cl_ctr) &&
								    j - 4 > load_op(op_ctr) &&
								    mask[j - 4] == 92 && mask[j - 3] == 'x' &&
								    isxdigit((int)mask[j - 2]) && isxdigit((int)mask[j - 1]) &&
								    mask[j + 1] == 92 && mask[j + 2] == 'x' &&
								    isxdigit((int)mask[j + 3]) && isxdigit((int)mask[j + 4])) {
									char str[3] = "xx";
									dash_loc_beg[k] = j - 4;
									dash_loc_end[k] = j + 4;
									str[0] = mask[j - 2];
									str[1] = mask[j - 1];
									start_c[k] = strtol(str, NULL, 16);
									str[0] = mask[j + 3];
									str[1] = mask[j + 4];
									end_c[k] = strtol(str, NULL, 16);
									}
								else {
									dash_loc_beg[k] = j - 1;
									dash_loc_end[k] = j + 1;
									start_c[k] = (int)mask[j - 1];
									end_c[k] = (int)mask[j + 1];
								}
								if (start_c[k] > end_c[k]) {
									int tmp;
									tmp = start_c[k];
									start_c[k] = end_c[k];
									end_c[k] = tmp;
								}
								k++;
							}
			}

			for (j = load_op(op_ctr) + 1, k = 0; j < load_cl(cl_ctr); j++) {
				if (j == dash_loc_beg[k]) {
					int ctr = start_c[k];
					while (ctr <= end_c[k]) {
						int t;
						/* Remove redundancy */
						for (t = 0; t < count(i); t++)
						      if((int)(cpu_mask_ctx->ranges[i].chars[t]) == ctr)
								goto cont1;
						cpu_mask_ctx->ranges[i].chars[count(i)++] = ctr;
						cont1: ;
						ctr++;
					}
					j = dash_loc_end[k];
					k++;
				}
				else {
					int t;
					/* Remove redundancy */
					for (t = 0; t < count(i); t++)
					      if((int)(cpu_mask_ctx->ranges[i].chars[t]) == mask[j])
								goto cont2;
					cpu_mask_ctx->ranges[i].chars[count(i)++] = mask[j];
					cont2: ;
				}
			}
			op_ctr++;
			cl_ctr++;
			cpu_mask_ctx->count++;
			fprintf(stdout, "[%d]\n", i);
		}
		else if((unsigned int)load_op(op_ctr) > (unsigned int)load_qtn(qtn_ctr))  {
			int j;
			cpu_mask_ctx->ranges[i].pos = calc_pos_in_key(mask, parsed_mask, load_qtn(qtn_ctr));

			switch(mask[load_qtn(qtn_ctr) + 1]) {
				case 'a': cpu_mask_ctx->ranges[i].start = 32;
					  for(j = 0; j < 95; j++)
						cpu_mask_ctx->ranges[i].chars[j] = 32 + j;
					  count(i) = 95;
					  break;
				case 'l': cpu_mask_ctx->ranges[i].start = 97;
					  for(j = 0; j < 26; j++)
						cpu_mask_ctx->ranges[i].chars[j] = 97 + j;
					  count(i) = 26;
					  break;
				case 'u': cpu_mask_ctx->ranges[i].start = 65;
					  for(j = 0; j < 26; j++)
						cpu_mask_ctx->ranges[i].chars[j] = 65 + j;
					  count(i) = 26;
					  break;
				case 'd': cpu_mask_ctx->ranges[i].start = 48;
					  for(j = 0; j < 10; j++)
						cpu_mask_ctx->ranges[i].chars[j] = 48 + j;
					  count(i) = 10;
					  break;
				case 's': for(j = 32; j <= 47; j++)
						cpu_mask_ctx->ranges[i].chars[count(i)++] = j;
				          for(j = 58; j <= 64; j++)
						cpu_mask_ctx->ranges[i].chars[count(i)++] = j;
					  for(j = 91; j <= 96; j++)
						cpu_mask_ctx->ranges[i].chars[count(i)++] = j;
					  for(j = 123; j <= 126; j++)
						cpu_mask_ctx->ranges[i].chars[count(i)++] = j;
					  break;
				default:  fprintf(stderr, "Unrecognized placeholder ?%c.\n", mask[load_qtn(qtn_ctr) + 1]);
					  exit(0);
			}
			qtn_ctr++;
			cpu_mask_ctx->count++;
			fprintf(stdout, "?%d\n", i);
		}
	}
#undef count
	for (i = 0; i < cpu_mask_ctx->count - 1; i++) {
		cpu_mask_ctx->ranges[i].next = i + 1;
		cpu_mask_ctx->active_positions[i] = 1;
	}
	cpu_mask_ctx->ranges[i].next = MAX_NUM_MASK_PLHDR - 1;
	cpu_mask_ctx->active_positions[i] = 1;
}

/* Returns the template of the keys corresponding to the mask.
 * Wordlist + mask not taken into account.
 */
static char* generate_template_key(char *mask, parsed_ctx *parsed_mask) {
	char *template_key = (char*)malloc(0x400);
	int i, k, t;
	int search_stack(parsed_ctx *parsed_mask, int loc) {
		int t;
		for (t = 0; load_op(t) != -1; t++)
			if(load_op(t) <= loc && load_cl(t) >= loc)
				return load_cl(t);
		for (t = 0; load_qtn(t) != -1; t++)
			if(load_qtn(t) == loc)
				return load_qtn(t) + 1;
		return 0;
	}
	i = 0, k = 0;
	while (i < strlen(mask)) {
		t = search_stack(parsed_mask, i);
		if (t){
			template_key[k++] = '#';
			i = t + 1;
		}
		else
			template_key[k++] = mask[i++];
	}
	template_key[k] = '\0';

	return template_key;
}

static void generate_keys(char *template_key, cpu_mask_context *cpu_mask_ctx, int my_words, int their_words) {
	int i, j, k, ps1 = MAX_NUM_MASK_PLHDR - 1, ps2 = MAX_NUM_MASK_PLHDR - 1, ps3 = MAX_NUM_MASK_PLHDR - 1, ps;
	int offset = cpu_mask_ctx->offset, num_active_postions = 0;

	for (i = 0; i < cpu_mask_ctx->count; i++)
		if((int)(cpu_mask_ctx->active_positions[i])) {
			ps1 = i;
			break;
		}

	ps2 = cpu_mask_ctx->ranges[ps1].next;
	ps3 = cpu_mask_ctx->ranges[ps2].next;

	for (i = 0; i < cpu_mask_ctx->count; i++)
		if((int)(cpu_mask_ctx->active_positions[i])) {
			num_active_postions++;
			cpu_mask_ctx->ranges[i].iter = 0;
		}

	if (!num_active_postions)
		fprintf(stdout, "%s\n", template_key);

	else if (num_active_postions == 1) {
		for (i = 0; i < cpu_mask_ctx->ranges[ps1].count; i++) {
			template_key[cpu_mask_ctx->ranges[ps1].pos + offset] =
			cpu_mask_ctx->ranges[ps1].chars[i];
			fprintf(stdout, "%s\n", template_key);
		}
	}

	else if (num_active_postions == 2) {
		for (j = 0; j < cpu_mask_ctx->ranges[ps2].count; j++) {
			template_key[cpu_mask_ctx->ranges[ps2].pos + offset] =
			cpu_mask_ctx->ranges[ps2].chars[j];
			for (i = 0; i < cpu_mask_ctx->ranges[ps1].count; i++) {
				template_key[cpu_mask_ctx->ranges[ps1].pos + offset] =
				cpu_mask_ctx->ranges[ps1].chars[i];
				fprintf(stdout, "%s\n", template_key);
			}
		}
	}

	else if (num_active_postions > 2) {

#define iter cpu_mask_ctx->ranges[ps].iter

		ps = cpu_mask_ctx->ranges[ps3].next;

		/* Initialize the reaming placeholders other than the first three */
		while(ps != MAX_NUM_MASK_PLHDR - 1) {
			template_key[cpu_mask_ctx->ranges[ps].pos + offset] =
			cpu_mask_ctx->ranges[ps].chars[0];
			ps = cpu_mask_ctx->ranges[ps].next;
		}

		while (1) {
			/* Iterate over first three placeholders */
			for (k = 0; k < cpu_mask_ctx->ranges[ps3].count; k++) {
				template_key[cpu_mask_ctx->ranges[ps3].pos + offset] =
				cpu_mask_ctx->ranges[ps3].chars[k];
				for (j = 0; j < cpu_mask_ctx->ranges[ps2].count; j++) {
					template_key[cpu_mask_ctx->ranges[ps2].pos + offset] =
					cpu_mask_ctx->ranges[ps2].chars[j];
					for (i = 0; i < cpu_mask_ctx->ranges[ps1].count; i++) {
						template_key[cpu_mask_ctx->ranges[ps1].pos + offset] =
						cpu_mask_ctx->ranges[ps1].chars[i];

						if (options.node_count) {
							seq++;
							if (their_words) {
							    their_words--;
							    continue;
							}
							if (--my_words == 0) {
							    my_words =
								options.node_max - options.node_min + 1;
							    their_words = options.node_count - my_words;
							}
						}
						if (ext_filter(template_key))
							if (crk_process_key(template_key))
								goto done;
					}
				}
			}

			ps = cpu_mask_ctx->ranges[ps3].next;

			/* Calculate next state of remaing placeholders, working
			 * principle similar to counters.
			 */
			while(1) {

				if (ps == MAX_NUM_MASK_PLHDR - 1) goto done;
				if ((++iter) == cpu_mask_ctx->ranges[ps].count) {
					iter = 0;
					template_key[cpu_mask_ctx->ranges[ps].pos + offset] =
					cpu_mask_ctx->ranges[ps].chars[(int)(iter)];
					ps = cpu_mask_ctx->ranges[ps].next;
				}
				else {
					template_key[cpu_mask_ctx->ranges[ps].pos + offset] =
					cpu_mask_ctx->ranges[ps].chars[(int)(iter)];
					break;
				}
			}
		}
		done: ;
#undef iter
	}
}
/* Skips iteration for postions stored in arr */
void skip_position(cpu_mask_context *cpu_mask_ctx, int *arr) {
	if (arr != NULL) {
		int k = 0;
		while (arr[k] >= 0 && arr[k] < cpu_mask_ctx->count) {
			int j, i, flag1 = 0, flag2 = 0;
			cpu_mask_ctx->active_positions[arr[k]] = 0;
			cpu_mask_ctx->ranges[arr[k]].next = MAX_NUM_MASK_PLHDR;

			for (j = arr[k] - 1; j >= 0; j--)
				if ((int)(cpu_mask_ctx->active_positions[j])) {
					flag1 = 1;
					break;
				}

			for (i = arr[k] + 1; i < cpu_mask_ctx->count; i++)
				if ((int)(cpu_mask_ctx->active_positions[i])) {
					flag2 = 1;
					break;
				}

			if (flag1)
				cpu_mask_ctx->ranges[j].next = flag2?i:MAX_NUM_MASK_PLHDR - 1;
			k++;
		}
	}
}



static double get_progress(void)
{
	double try;

	emms();

	try = ((unsigned long long)status.cands.hi << 32) + status.cands.lo;

	if (!cand)
		return -1;

	return 100.0 * try / cand;
}

static void save_state(FILE *file)
{
	int i;

	fprintf(file, "%u\n", rec_seq);
	fprintf(file, "%s\n", rec_ctx.output);
	fprintf(file, "%d\n", rec_ctx.count);
	for (i = 0; i < rec_ctx.count; i++)
		fprintf(file, "%d\n", rec_ctx.ranges[i].index);
}

static int restore_state(FILE *file)
{
	int i;

	if (fscanf(file, "%u\n", &seq) != 1)
		return 1;
	if (fscanf(file, "%128[^\n]\n", ctx.output) != 1)
		return 1;
	if (fscanf(file, "%d\n", &ctx.count) != 1)
		return 1;
	for (i = 0; i < ctx.count; i++)
		if (fscanf(file, "%d\n", &ctx.ranges[i].index) != 1)
			return 1;
	return 0;
}

static void fix_state(void)
{
	rec_seq = seq;
	rec_ctx = ctx;
}

void do_mask_crack(struct db_main *db, char *mask)
{
	int my_words, their_words;
	int i;

	/* We do not yet support min/max-len */
	if (options.force_minlength >= 0 || options.force_maxlength) {
		fprintf(stderr, "Mask mode: --min-length and --max-length currently not supported\n");
		error();
	}

	log_event("Proceeding with mask mode");

	parse_braces(mask, &parsed_mask);
	if (parsed_mask.parse_ok)
		parse_qtn(mask, &parsed_mask);
	else {
		fprintf(stderr, "Parsing unsuccessful\n");
		error();
	}

	init_cpu_mask(mask, &parsed_mask, &cpu_mask_ctx);

	/*
	 * Warning: NULL to be raplaced by an array containing information
	 * regarding GPU portion of mask.
	 */
	skip_position(&cpu_mask_ctx, NULL);

	seq = 0;

	status_init(&get_progress, 0);


	rec_restore_mode(restore_state);
	rec_init(db, save_state);

	crk_init(db, fix_state, NULL);

	my_words = options.node_max - options.node_min + 1;
	their_words = options.node_min - 1;

	if (seq) {
/* Restored session.  seq is right after a word we've actually used. */
		int for_node = seq % options.node_count + 1;
		if (for_node < options.node_min ||
		        for_node > options.node_max) {
/* We assume that seq is at the beginning of other nodes' block */
			their_words = options.node_count - my_words;
		} else {
			my_words = options.node_max - for_node + 1;
			their_words = 0;
		}
	}

	cand = 1;
	for (i = 0; i < cpu_mask_ctx.count; i++)
		if ((int)(cpu_mask_ctx.active_positions[i]))
			cand *= cpu_mask_ctx.ranges[i].count;
	if (options.node_count)
		cand *= (double)(options.node_max - options.node_min + 1) /
			options.node_count;

	/*while(!crk_process_key("a"));*/
/*
	while ((word = rpp_next(&ctx))) {
		if (options.node_count) {
			seq++;
			if (their_words) {
				their_words--;
				continue;
			}
			if (--my_words == 0) {
				my_words =
					options.node_max - options.node_min + 1;
				their_words = options.node_count - my_words;
			}
		}
		if (ext_filter(word))
			if (crk_process_key(word))
				break;
	}*/

	generate_keys(generate_template_key(mask, &parsed_mask), &cpu_mask_ctx, my_words, their_words);

	// For reporting DONE regardless of rounding errors
	if (!event_abort)
		cand = ((unsigned long long)status.cands.hi << 32) +
			status.cands.lo;

	crk_done();

	rec_done(event_abort);
}
